from pyadlml.dataset._representations.raw import create_raw 
from pyadlml.dataset._representations.changepoint import create_changepoint
from pyadlml.dataset._representations.lastfired import create_lastfired
from pyadlml.dataset._representations.image import create_lagged_raw, create_lagged_lastfired, \
                                            create_lagged_changepoint
from pyadlml.dataset._dataset import label_data
from pyadlml.dataset.devices import device_rep1_2_rep2
import sklearn.preprocessing as preprocessing
import pandas as pd
import numpy as np

REPS = ['raw', 'lastfired', 'changepoint']

class DiscreteEncoder():
    def __init__(self, rep, t_res=None, sample_strat='ffill'):
        assert rep in REPS

        self.rep = rep
        self.t_res = t_res
        self.sample_strat = sample_strat
        self.data = None

    def fit(self, df_devices):
        if self.rep == 'raw':
            self.data = create_raw(
                df_devices,
                t_res=self.t_res,
                sample_strat=self.sample_strat
            )
        elif self.rep == 'changepoint':
            self.data = create_changepoint(
                df_devices,
                t_res=self.t_res
            )
        elif self.rep == 'lastfired':
            self.data = create_lastfired(
                df_devices,
                t_res=self.t_res
            )
        else:
            raise ValueError


    def fit_transform(self, df_devices):
        self.fit(df_devices)
        return self.data

    def inverse_transform(self, raw):
        """
        """
        raise NotImplementedError

    def set_params(self, t_res=None, sample_strat=None):
        if t_res is not None:
            self.t_res = t_res
        if sample_strat is not None:
            self.sample_strat = sample_strat

    def transform(self, df_devices=None):
        if self.rep == 'raw':
            return create_raw(
                df_devices,
                t_res=self.t_res,
                sample_strat=self.sample_strat
            )
        elif self.rep == 'changepoint':
            return create_changepoint(
                df_devices,
                t_res=self.t_res
            )
        elif self.rep == 'lastfired':
            return create_lastfired(
                df_devices,
                t_res=self.t_res
            )

class LabelEncoder():
    """
        wrapper around labelencoder to handle time series data
    """
    def __init__(self, df_devices, idle=False):
        self.labels = None
        self.df_devices = df_devices
        self.idle = idle
        self._lbl_enc = preprocessing.LabelEncoder()

    def fit(self, df_activities):
        """ labels data and creates the numeric representations 
        Parameters
        ----------
        df_activities : pd.DataFrame
            Columns are end_time, start_time, activity. 
        """
        df = label_data(self.df_devices, df_activities, self.idle)
        self._lbl_enc.fit(df['activity'].values)

    def fit_transform(self, df_activities):
        df = label_data(self.df_devices, df_activities, self.idle)
        encoded_labels = self._lbl_enc.fit_transform(df['activity'].values)
        return pd.DataFrame(index=df.index, data=encoded_labels, columns=['activity'])

    def inverse_transform(self, x, retain_index=False):
        """
        Parameters
        ----------
        x: array like or pd.DataFrame or pd.Series
            array of numbers that are transformed to labels
        """
        if isinstance(x, np.ndarray):
            res = self._lbl_enc.inverse_transform(x)
            if retain_index:
                return pd.Series(data=res, index=self.df_devices.index[:len(res)])
            else:
                return res
                
        elif isinstance(x, pd.DataFrame) or isinstance(x, pd.Series): 
            tmp_index = x.index
            res = self._lbl_enc.inverse_transform(x.values)
            return pd.Series(data=res, index=tmp_index)
        else:
            raise ValueError

    def set_params(self, t_res=None, sample_strat=None):
        if t_res is not None:
            self.t_res = t_res
        if sample_strat is not None:
            self.sample_strat = sample_strat

    def get_params(self):
        return self.labels, self.idle, self.df_devices

    def transform(self, x):
        """
        
        """
        # if the input is a dataframe of activities, than fit and
        # transform the data accordingly 
        col_names = {'start_time', 'end_time', 'activity'}
        if isinstance(x, pd.DataFrame) and set(x.columns) == col_names:
            df = label_data(self.df_devices, x, self.idle)
            encoded_labels = self._lbl_enc.transform(df['activity'].values)
            return pd.DataFrame(index=df.index, data=encoded_labels, columns=['activity'])
        
        # return only the labels for a nd array 
        elif isinstance(x , np.ndarray):
            return self._lbl_enc.transform(x)
        else:
            raise ValueError


class ImageEncoder():
    def __init__(self, rep, window_size, t_res=None, sample_strat='ffill'):
        self.data = None
        self.rep = rep
        self.t_res = t_res
        self.window_size = window_size
        self.sample_strat = sample_strat

    def fit(self, df_devices):
        if self.rep == 'raw':
            self.data = create_lagged_raw(
                df_devices, 
                window_size=self.window_size, 
                t_res=self.t_res, 
                sample_strat=self.sample_strat)

        elif self.rep == 'changepoint':
            self.data = create_lagged_changepoint(
                df_devices, 
                window_size=self.window_size, 
                t_res=self.t_res)

        elif self.rep == 'lastfired':
            return create_lagged_lastfired(
            df_devices, 
            window_size=self.window_size, 
            t_res=self.t_res)

    def fit_transform(self, df_devices):
        self.fit(df_devices)
        return self.data

    def inverse_transform(self, lgd_raw):
        """
        """
        raise NotImplementedError

    def set_params(self, t_res=None, window_size=10, sample_strat='ffill'):
        if t_res is not None:
            self.t_res = t_res
        raise NotImplementedError

    def transform(self, df_devices):
        if self.rep == 'raw':
            return create_lagged_raw(
                df_devices, 
                window_size=self.window_size, 
                t_res=self.t_res, 
                sample_strat=self.sample_strat)
        elif self.rep == 'changepoint':
            return create_lagged_changepoint(
                df_devices, 
                window_size=self.window_size, 
                t_res=self.t_res)
        elif self.rep == 'lastfired':
            return create_lagged_lastfired(
            df_devices, 
            window_size=self.window_size, 
            t_res=self.t_res)
        else:
            raise ValueError


class ImageLabelEncoder():
    """
    wrapper around labelencoder to handle time series data
    """
    def __init__(self, df_devices, window_size, t_res=None, idle=False):
        self.window_size = window_size
        self.t_res = t_res
        self.idle = idle
        self._lbl_enc = preprocessing.LabelEncoder()
        self.df_devices = df_devices
        self.df_index = self._create_index(df_devices, t_res)
        
    def _create_index(self, df_devices, t_res):
        """
        create the dummy dataframe for the index from the devices
        index | val
        """
        df = df_devices.copy()
        df = df.pivot(index='time', columns='device', values='val').iloc[:,:1]
        df = df.astype(bool) # just to have a lower memory footprint
        
        # resample with frequency
        resampler = df.resample(t_res, kind='timestamp')
        df_index = resampler.sum()
        df_index.columns = ['val']
        df_index['val'] = 1
        return df_index

        
    def fit(self, df_activities):
        df = label_data(self.df_index, df_activities, self.idle)
        # start where the labeling begins
        df = df.iloc[self.window_size:,:]
        self._lbl_enc.fit(df['activity'].values)

    def fit_transform(self, df_activities):
        df = label_data(self.df_index, df_activities, self.idle)
        df = df.iloc[self.window_size:,:]
        encoded_labels = self._lbl_enc.fit_transform(df['activity'].values)
        return pd.DataFrame(index=df.index, data=encoded_labels, columns=['activity'])

    def inverse_transform(self, x, retain_index=False):
        """
        Parameters
        ----------
        x: array like or pd.DataFrame or pd.Series
            array of numbers that are transformed to labels
        """
        if isinstance(x, np.ndarray):
            res = self._lbl_enc.inverse_transform(x)
            if retain_index:
                return pd.Series(data=res, index=self.df_index.index[:len(res)])
            else:
                return res
                
        elif isinstance(x, pd.DataFrame) or isinstance(x, pd.Series): 
            tmp_index = x.index
            res = self._lbl_enc.inverse_transform(x.values)
            return pd.Series(data=res, index=tmp_index)
        else:
            raise ValueError

    def set_params(self, t_res=None, idle=None, window_size=None):
        if t_res is not None:
            self.t_res = t_res
            self.df_index = self._create_index(self.df_devices, t_res)
        if idle is not None:
            self.idle = idle
        if window_size is not None:
            self.window_size = window_size

    def get_params(self):
        return {'window_size': self.window_size, 'idle': self.idle, 't_res':self.t_res}

    def transform(self, df_activities):
        df = label_data(self.df_index, df_activities, self.idle)
        df = df.iloc[self.window_size:,:]
        encoded_labels = self._lbl_enc.transform(df['activity'].values)
        return pd.DataFrame(index=df.index, data=encoded_labels, columns=['activity'])




import numpy as np
from datetime import timedelta
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import matplotlib.colors as colors

from pyadlml.dataset.stats.devices import duration_correlation, \
    trigger_time_diff, device_tcorr, device_triggers_one_day, \
    devices_trigger_count, devices_on_off_stats

from pyadlml.dataset.plot.util import heatmap_square, hm_key_NN, \
    func_formatter_sec, heatmap, annotate_heatmap, savefig

from pyadlml.dataset.devices import _is_dev_rep2, device_rep1_2_rep2

DEV_BAR_HM = {10:(9,5), 12:(10,6), 14: (10,6), 20: (10,9), 22:(10,9),40:(10,15), 72:(10,21)}
DEV_BP_HM = {12:(9,6), 20:(10,9), 22:(10,8), 72:(11,22)}
DEV_HM_HM = {12:(8,8), 20:(9,9), 22:(13,13), 40:(15,15),72:(23,22)}
DEV_HM_TIME = {12:(10,6), 20:(10,8)}

def num_device_2_figsize(hm, figsize, df):
    if figsize is not None:
        return figsize
    if isinstance(df, pd.DataFrame):
        num_dev = len(df['device'].unique())
    elif isinstance(df, list):
        num_dev = len(df)
    else:
        raise ValueError
    return hm_key_NN(hm, num_dev)

def hist_trigger_time_diff(df_dev=None, x=None, n_bins=50, figsize=(10,6), file_path=None):
    """
        plots
    """
    assert not (df_dev is None and x is None)
    title='Time difference between device triggers'
    log_sec_col = 'total_log_secs'
    sec_col = 'total_secs'

    if x is None:
        X = trigger_time_diff(df_dev.copy())
    else:
        X = x

    # make equal bin size from max to min
    bins = np.logspace(min(np.log10(X)), max(np.log10(X)), n_bins)

    # make data ready for hist
    hist, _ = np.histogram(X, bins=bins)
    cum_percentage = hist.cumsum()/hist.sum()
    cum_percentage = np.concatenate(([0], cum_percentage)) # let the array start with 0

    # plots
    fig,ax = plt.subplots(figsize=figsize)
    plt.xscale('log')
    #plt.yscale('log')
    
    ax.hist(X, bins=bins, label='amount of trigger with set difference')
    ax.set_ylabel('count')
    ax.set_xlabel('log seconds')
    
    # create axis for line
    ax2=ax.twinx()
    ax2.plot(bins, cum_percentage, 'r', label='% of data to the left')
    ax2.set_ylabel('%')
    ax2.set_xscale('log')
    
    ax_top = ax.secondary_xaxis('top', functions=(lambda x: x, lambda x: x))
    ax_top.xaxis.set_major_formatter(
        ticker.FuncFormatter(func_formatter_sec))
    
    # plot single legend for multiple axis
    h1, l1 = ax.get_legend_handles_labels()
    h2, l2 = ax2.get_legend_handles_labels()
    ax.legend(h1+h2, l1+l2, loc='center right')
    
    plt.title(title, y=1.08)
    
    if file_path is not None:
        savefig(fig, file_path)
        return 
    else:
        return fig

def boxplot_on_duration(df_dev, figsize=None, file_path=None):
    """
    draws a boxplot of all devices
    Parameters
    ----------
        df_dev: pd.DataFrame
            device dataframe in representation 1 
    """
    title = 'Devices on-duration'
    figsize = num_device_2_figsize(DEV_BP_HM, figsize, df_dev)    
    xlabel = 'log seconds'
    xlabel_top = 'time'

    if not _is_dev_rep2(df_dev):
        df_dev = device_rep1_2_rep2(df_dev.copy())

    # create duration differences
    df_dev = df_dev.copy()
    df_dev['td'] = df_dev['end_time'] - df_dev['start_time']

    # select data for each device
    devices = df_dev['device'].unique()
    dat = []
    for device in devices:
        df_device = df_dev[df_dev['device'] == device]
        #tmp = np.log(df_device['td'].dt.total_seconds())
        tmp = df_device['td'].dt.total_seconds()
        dat.append(tmp)
    
    #return df_dev #DEBUG
    # plot boxsplot
    fig, ax = plt.subplots(figsize=figsize)
    ax.boxplot(dat, vert=False)
    ax.set_title(title)
    ax.set_yticklabels(devices, ha='right')
    ax.set_xlabel(xlabel)
    ax.set_xscale('log')

    # create secondary axis with 

    # create secondary axis with time format 1s, 1m, 1d
    ax_top = ax.secondary_xaxis('top', functions=(lambda x: x, lambda x: x))
    ax_top.set_xlabel(xlabel_top)
    ax_top.xaxis.set_major_formatter(
        ticker.FuncFormatter(func_formatter_sec))
    if file_path is not None:
        savefig(fig, file_path)
        return 
    else:
        return fig

def heatmap_trigger_one_day(df_dev=None, df_tod=None, t_res='1h', figsize=None, file_path=None):
    """
    computes the heatmap for one day where all the device triggers are showed
    """
    assert not (df_dev is None and df_tod is None)
    title = "Device triggers cummulative over one day"
    xlabel =  'time'
    figsize = num_device_2_figsize(DEV_HM_TIME, figsize, df_dev)

    if df_tod is None:
        df = device_triggers_one_day(df_dev.copy(), t_res)
    else:
        df = df_tod

    x_labels = list(df.index)
    y_labels = df.columns
    dat = df.values.T
    
    # begin plotting
    fig, ax = plt.subplots(figsize=figsize)
    im, cbar = heatmap(dat, y_labels, x_labels, ax=ax, cmap='viridis', cbarlabel='counts')
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    
    # format the x-axis
    def func(x,p):
        if True:
            if int(x/k) < 10:
                return '0{}:00'.format(int(x/k)+1)
            else:
                return '{}:00'.format(int(x/k)+1)
    
    # calculate the tick positions 
    a,b = ax.get_xlim()
    k = (b-a)/24
    tcks_pos = np.arange(0,23)*k + (-0.5 + k)
    
    x_locator = ticker.FixedLocator(tcks_pos)
    ax.xaxis.set_major_formatter(ticker.FuncFormatter(func))
    ax.xaxis.set_major_locator(x_locator)
    ax.set_aspect(aspect='auto')

    if file_path is not None:
        savefig(fig, file_path)
        return 
    else:
        return fig

def heatmap_trigger_time(df_dev=None, df_tcorr=None, t_window='5s', figsize=None, z_scale=None, file_path=None):
    """
    """
    assert not (df_dev is None and df_tcorr is None)
    title = "Triggercount with sliding window of " + t_window

    color = 'trigger count'
    cbarlabel = 'counts'
    figsize = num_device_2_figsize(DEV_HM_HM, figsize, df_dev)

    if df_tcorr is None:
        df = device_tcorr(df_dev, t_window)[0]
    else:
        df = df_tcorr

    # get the list of cross tabulations per t_window
    vals = df.astype(int).values.T
    devs = list(df.index)


    fig, ax = plt.subplots(figsize=figsize)

    log = True if z_scale == 'log' else False       
    valfmt = "{x:.0f}"
        
    im, cbar = heatmap_square(vals, devs, devs, ax=ax, #cmap='viridis', 
                        cbarlabel=cbarlabel, log=log)#, cbar_kw=cbar_kw)
    
    texts = annotate_heatmap(im, textcolors=("white", "black"), log=log, valfmt=valfmt)

    ax.set_title(title)
    fig.tight_layout()

    if file_path is not None:
        savefig(fig, file_path)
        return 
    else:
        return fig

def heatmap_cross_correlation(df_dev=None, df_dur_corr=None, figsize=None, parallel=True, numbers=True, file_path=None):
    """ plots the cross correlation between the device signals
    Parameters
    ----------
    df_dev: pd.DataFrame 
        devices in representation 1
    """
    assert not (df_dev is None and df_dur_corr is None)

    title = 'Cross-correlation of signals'
    cmap = 'BrBG'
    cbarlabel = 'counts'
    
    if df_dur_corr is None:
        if parallel:
            from pyadlml.dataset.stats.devices import duration_correlation_parallel
            ct = duration_correlation_parallel(df_dev)
        else:
            ct = duration_correlation(df_dev)
    else:
        ct = df_dur_corr 
    vals = ct.values.T
    devs = list(ct.index)

    figsize = num_device_2_figsize(DEV_HM_HM, figsize, devs)
    fig, ax = plt.subplots(figsize=figsize)
    im, cbar = heatmap_square(vals, devs, devs, ax=ax, cmap=cmap, cbarlabel=cbarlabel,
                       vmin=-1, vmax=1)
    if numbers:
        texts = annotate_heatmap(im, textcolors=("black", "white"), 
                             threshold=0.5, valfmt="{x:.2f}")

    ax.set_title(title)
    fig.tight_layout()
    plt.show()

def hist_on_off(df_dev=None, df_onoff=None, figsize=None, file_path=None):
    """ plots the percentage a device is on against the percentage it is off
    Parameters
    ----------
    df_dev: Dataframe
        device list in representation 1
    """
    assert not (df_dev is None and df_onoff is None)

    title = 'Devices fraction on/off'
    xlabel ='Percentage in binary states' 
    ylabel = 'Devices'

    figsize = num_device_2_figsize(DEV_BAR_HM, figsize, df_dev)

    if df_dev is not None and not _is_dev_rep2(df_dev):
        df_dev = device_rep1_2_rep2(df_dev.copy())

    if df_onoff is None:
        df = devices_on_off_stats(df_dev)
    else:
        df = df_onoff
    df = df.sort_values(by='frac_on', axis=0)
    dev_lst = list(df.index)
    # Figure Size 
    fig, ax = plt.subplots(figsize=figsize) 
    plt.barh(dev_lst, df['frac_off'].values, label='off')  

    # careful: notice "bottom" parameter became "left"
    plt.barh(dev_lst,  df['frac_on'].values, left=df['frac_off'], label='on')

    # we also need to switch the labels
    plt.title(title)
    plt.xlabel(xlabel)  
    plt.ylabel(ylabel)
    
    # set 
    widths = df['frac_off']
    xcenters = widths/2
    text_color='white'
    for y, (x, c) in enumerate(zip(xcenters, widths)):
        ax.text(x, y, '{:.4f}'.format(c), ha='center', va='center',
                color=text_color)
    
    ax.legend(ncol=2, bbox_to_anchor=(0, 1),
              loc='upper left', fontsize='small')
    
    # Remove axes splines 
    for s in ['top', 'right']: 
        ax.spines[s].set_visible(False) 

    plt.show()  

def hist_counts(df_dev=None, df_tc=None, figsize=None, y_scale=None, file_path=None):
    """
    plots the trigger count of each device 
    """
    assert not (df_dev is None and df_tc is None)
    title = 'Count of on/off activations per Device'
    col_label = 'trigger count'
    col_device = 'device'

    figsize = num_device_2_figsize(DEV_BAR_HM, figsize, df_dev)

    if df_tc is None:
        df = devices_trigger_count(df_dev.copy())
    else:
        df = df_tc

    df.reset_index(level=0, inplace=True)
    df.columns = ['device', col_label]

    df = df.sort_values(by=col_label, axis=0, ascending=True)
    
    # plot
    fig, ax = plt.subplots(figsize=figsize)
    plt.title(title)
    plt.xlabel(col_label)
    ax.barh(df['device'], df['trigger count'])
    if y_scale == 'log':
        ax.set_xscale('log')

    if file_path is not None:
        savefig(fig, file_path)
        return 
    else:
        return fig


from pyadlml.dataset import START_TIME, ACTIVITY, END_TIME
import pandas as pd
import numpy as np

"""
    df_activities:
        - per definition no activity can be performed in parallel

        start_time | end_time   | activity
        ---------------------------------
        timestamp   | timestamp | act_name


"""

def check_activities(df):
    """
    check if the activitiy dataframe is valid by checking if
        - the dataframe has the correct dimensions and labels
        - activities are non overlapping

    :param: df 
    start_time | end_time   | activity
    ----------------------------------
    timestamp   | timestamp | act_name

    """
    if not START_TIME in df.columns or not END_TIME in df.columns \
    or not ACTIVITY in df.columns or len(df.columns) != 3:
        print('the lables and dimensions of activites does not fit')
        raise ValueError

    if _is_activity_overlapping(df):
        print('there should be none activity overlapping')
        raise ValueError
    return True

def _is_activity_overlapping(df):
    """ checks if any activity is overlapping another
    Parameters
    ----------
    df : pd.DataFrame
    
    """
    assert df.shape[1] == 3
    epsilon = pd.Timedelta('0ms')
    mask = (df[END_TIME].shift()-df[START_TIME]) > epsilon
    overlapping = df[mask]
    return not overlapping.empty

def _create_activity_df():
    """
    returns: empty pd Dataframe 
    """
    df = pd.DataFrame(columns=[START_TIME, END_TIME, ACTIVITY])
    df[START_TIME] = pd.to_datetime(df[START_TIME])
    df[END_TIME] = pd.to_datetime(df[END_TIME])
    return df 


def add_idle(acts, min_diff=pd.Timedelta('5s')):
    """ adds a dummy Idle activity for gaps between activities greater than min_diff
    Parameters
    ----------
    acts: pd.DataFrame
        the activity data with columns: start_time, end_time, activity
    Returns
    ----------
    pd.DataFrame
        activity data with columns: start_time, end_time, activity
    """
    acts = acts.copy().reset_index(drop=True)
    def func(series, df):
        """
        """
        # check if at end of the series
        if series.name == len(df)-1:
            return series
        else:
            next_entry = df.loc[series.name+1,:]
            return pd.Series({
                'start_time': series.end_time + pd.Timedelta('1ms'),
                'end_time' : next_entry.start_time - pd.Timedelta('1ms'),
                'activity' : 'idle'
            })                

     
    acts['diff'] =  acts['start_time'].shift(-1) - acts['end_time']
    tmp = acts[acts['diff'] > min_diff].copy()
    tmp = tmp.apply(func, axis=1, df=acts)
    res = pd.concat([acts.iloc[:,:-1],tmp]).sort_values(by='start_time')
    return res.reset_index(drop=True)


def _merge_int_inclusive(row, ov, strat='inplace'):
    """
    int |~~~~~~~~|
    ov    |----|   => |~|----|~~|
    
    Parameters
    ----------
    ov: pd.DataFrame
        single df with one row
    row: pd.Series
        one row 
    """
    assert strat in ['inplace']
    # use epsilon to offset the interval boundaries a little bit
    # to prevent later matching of multiple indices
    eps = pd.Timedelta('1ms')
    df = _create_activity_df()
    
    if strat == 'inplace':
        df.loc[0] = [row.start_time, ov.start_time, row.activity]
        df.loc[1] = [ov.start_time + eps, ov.end_time, ov.activity]
        df.loc[2] = [ov.end_time + eps, row.end_time, row.activity]
    else:
        raise ValueError #Should never happen
    return df

def _merge_int_right_partial(row, ov, strat='clip_left'):
    """ merges 
        row int |~~~~|    => |~~~|---|    
        ov ints   |----|  
    """
    df_res = _create_activity_df()
    eps = pd.Timedelta('1ms')
    
    # get row stats
    row_act = row[ACTIVITY]
    row_st = row[START_TIME]
    row_et = row[END_TIME]
    
    # get overlap stats
    ov_st = ov.start_time
    ov_et = ov.end_time
    ov_act = ov.activity
    
    if strat == 'clip_left':
        df_res.loc[0] = [row_st, ov_st, row_act]
        df_res.loc[1] = [ov_st + eps, ov_et, ov_act]
    return df_res

def _merge_int_first_persists(row1, row2):
    """ replaces interval by taking the first as dominant interval which borders 
        should be preserved:
            1. case 
                row1 int |~~~~|    => |~~~~|-|    
                row2 int   |----|  
            2. case
                row1 int   |~~~~|  => |-|~~~~|    
                row2 int |----|
    Parameters
    ----------
    row1 : pd.Series
        The dominant interval
    row2 : pd.Series
        the submissive interval
    Returns
    -------
    df_res : pd.DataFrame
        the corrected activities
    """
    df = _create_activity_df()
    eps = pd.Timedelta('1ms')

    int1 = pd.Interval(row1.start_time, row1.end_time)
    int2 = pd.Interval(row2.start_time, row2.end_time)

    if int1.left <= int2.left and int1.right < int2.right: # 1.case
        df.loc[0] = [row1.start_time, row1.end_time, row1.activity]
        df.loc[1] = [row1.end_time + eps, row2.end_time, row2.activity]

    elif int2.left <= int1.left and int2.right < int1.right: # 2.case
        df.loc[0] = [row2.start_time, row1.start_time, row2.activity]
        df.loc[1] = [row1.start_time + eps, row1.end_time, row1.activity]
    return df

def _merge_int_same(row1, row2):
    """ the intervals are the same  
    Parameters
    ----------
    row1 : pd.Series
    row2 : pd.Series
    Returns
    -------
    df_res : pd.DataFrame
        the corrected activities
    """
    assert row1.activity == row2.activity
    
    df = _create_activity_df()
    df.loc[0] = [row1.start_time, row2.end_time, row1.activity]
    return df



def _merge_ints(row, overlapping, strats=['cut_at_lower'], excepts=[]):
    """ gets overlapping intervals and merges those intervals with a strategy
    Parameters
    ----------
    row : pd.Series 

    overlapping : pd.Series 

    strats : list
        the strategies for merging intervals with priority in ascending order
    excepts : list
        the activities which should be preserved when merging
 
    Returns
    -------
    merged: pd.DataFrame
    """
    # todo check if every strategy in the list is possible
    assert isinstance(overlapping, pd.Series)
    assert isinstance(row, pd.Series)

    if row.activity == overlapping.activity:
        return _merge_int_same(row, overlapping)
        
    int1 = pd.Interval(row.start_time, row.end_time)
    int2 = pd.Interval(overlapping.start_time, overlapping.end_time)
    
    if excepts != []:
        # find out which is the dominant and the less dominant interval
        if row.activity in excepts and overlapping.activity in excepts:
            # assign row and ov such as to replace the one with lower priority below
            print('priority mismatch!'*20)
            idx_row = excepts.index(row.activity)
            idx_ov = excepts.index(overlapping.activity)
            if idx_row < idx_ov:
                dom = overlapping
                less_dom = row

        if row.activity in excepts:
            dom = row
            less_dom = overlapping
        else:
            dom = overlapping 
            less_dom = row


        # apply merging strategies
        int_dom = pd.Interval(dom.start_time, dom.end_time)
        int_ldom = pd.Interval(less_dom.start_time, less_dom.end_time)

        # dominant interval encloses less dominant => keep dominant, drop less dominant
        if (int_dom.left < int_ldom.left) & (int_ldom.right < int_dom.right):
            df_res = _create_activity_df()
            df_res.loc[0] = dom
            return df_res

        # less dominant interval encloses dominant => normal inclusive merge
        elif (int_ldom.left < int_dom.left) & (int_dom.right < int_ldom.right):
            return _merge_int_inclusive(less_dom, dom)

        # intervals overlap => keep dominant
        else:
            return _merge_int_first_persists(dom, less_dom)
            

    if (int1.left < int2.left) & (int2.right < int1.right):
        # int1  |~~~~~~|  
        # int2   |----|  
        df_res = _merge_int_inclusive(row, overlapping)
        
    elif (int1.left <= int2.left) & (int1.right < int2.right):
        # int1 |~~~~|   
        # int2   |----|   
        df_res = _merge_int_right_partial(row, overlapping)
        
    else:
        raise ValueError # this should never happen
    return df_res


def _is_activity_et_after_st(df):
    """ checks if the start_time is lesser than end_time for every activity
    """
    df = df.sort_values(by=START_TIME)
    df = df.copy()
    
    df['diff'] = df['end_time'] - df['start_time']
    mask = df['diff'] < pd.Timedelta('0ns')
    return not mask.any()



def correct_activity_overlap(df, strats=[], excepts=[]):
    """ solve the merge overlapping interval problem
        worst runtime is O(n^2)
        average runtime is O()
        
    Parameters
    ----------
    df : pd.DataFrame
        Activity dataframe with the columns
    strats : list
        the strategies for merging intervals with priority in ascending order
    excepts : list
        the activities which should be preserved when merging
 
    Returns
    -------
    df : pd.DataFrame 
        corrected activity dataframe
    corrections : list
        a list of tuples with the areas that had to be corrected and the corrections
    """
    df = df.copy()
    df = df.sort_values(START_TIME)
    df = df.reset_index(drop=True)
    
    res = _create_activity_df()
    corrections = [] 

    # get all activities that have in an overlap with a direct preceding interval
    mask = (df[START_TIME].shift(-1) - df[END_TIME]) < pd.Timedelta('0ms')
    idxs_succ_overlaps = np.array(list(df[mask].index))

    i_l, i = 0, 0
    while i < len(idxs_succ_overlaps):
        i_h = idxs_succ_overlaps[i]
        res = res.append(df.iloc[i_l:i_h,:]) # exclusive i_h e,g 0->2 is [0,1]
        
        # get index of first element that start_time is lesser than the end time
        # this marks the point where we can copy again indices
        i_l = i_h
        i_h = _get_idx_start(df, i_h)
        
        # only for last iteration
        if i == len(idxs_succ_overlaps)-1:
            area_to_correct = df.iloc[i_l:i_h,:]
            result = _correct_overlapping_activities(area_to_correct, strats, excepts)
            corrections.append((area_to_correct, result))
            res = res.append(result)
            break
            
        # if the index exceed the any pair of overlapping indicies, extent the range to
        # include them
        try:
            idxs_succ_overlaps2 = idxs_succ_overlaps[i+1:].copy()
        except:
            raise ValueError
            
        if (i_h >= idxs_succ_overlaps2).any():       
            """ 
            for all succ ov indicies get the maximum of the extented range
            e.g max-int = 4
               1 |~~~~~~~~~~~~~~|
               2          |--|
               3              |~~~~| 
               4                 |---|
               5                       |---| <----- max-int = 5
               6                            |~~~~|
               7                              |-|
            """
            for idx in idxs_succ_overlaps2:     
                # if the succ pair is in range of the overlap skip in next big iteration
                if idx < i_h: 
                    i += 1
                else:
                    break
                # set the upper overlap limit to the maximum of the included rows
                i_h = max(_get_idx_start(df, idx), i_h)
        
        area_to_correct = df.iloc[i_l:i_h,:]
        result = _correct_overlapping_activities(area_to_correct, strats, excepts)

        assert not result.empty
        corrections.append((area_to_correct, result))        
        res = res.append(result)        
        i += 1
        i_l = i_h
        
        
    res = res.append(df.iloc[i_h:,:])
    
    # sanity checks
    assert len(res) >= len(df)
    added_entrys = 0
    for corr in corrections:
        added_entrys += len(corr[1]) - len(corr[0])
    assert len(res) - added_entrys == len(df)
    
    res = res.sort_values(by=START_TIME)
    res = res.reset_index(drop=True)    
    return res, corrections

def _get_idx_start(df, idx):
    """ returns the index of the first element in the dataframe where the 
    start_time is greater than the end_time of the interval that overlaps 
    the succeding one. Thus marking the overlapping areas that need correction.
    Parameters
    ----------
    df : pd.DataFrame
        activity dataframe
    idx : int
        index of the row where the interval overlaps the succeding row(s)
    
    Returns
    -------
    res : int
        index of a row
    """
    et_of_violation = df.iloc[idx,:].end_time
    tmp = df[df[START_TIME] > et_of_violation]
    res = list(tmp.index)[0]
    return res

def _correct_overlapping_activities(area_to_correct, strats, excepts=[]):
    """
    Parameters
    ----------
    area_to_correct : pd.DataFrame
        Acitity dataframe
    strats : list
        the strategies for merging intervals with priority in ascending order
    excepts : list
        the activities which should be preserved when merging
     
    """
    assert len(area_to_correct) >= 2
    
    result = _create_activity_df()
    stack = area_to_correct.copy().reset_index(drop=True)
    
    while True:
        # pop first and second item from stack if they overlap otherwise 
        # append to result until two items overlap
        while True:                
            current_row = stack.iloc[0]
            ov = stack.iloc[1]

            # if they don't overlap push onto result otherwise proceed with merging 
            # procedure
            int1 = pd.Interval(current_row.start_time, current_row.end_time)
            int2 = pd.Interval(ov.start_time, ov.end_time)
            if int1.overlaps(int2):
                stack = stack.iloc[2:]
                break
            # the case when the last two activities on stack don't overlap
            elif stack.iloc[2:].empty:
                result = result.append(stack)
                return result
            else:
                result = result.append(current_row)
                stack = stack.iloc[1:]

 
        new_rows = _merge_ints(current_row, ov, strats, excepts)

        if stack.empty: 
            result = result.append(new_rows)
            return result
        elif len(new_rows) >= 2:
            result = result.append(new_rows.iloc[0,:])
            new_rows = new_rows.iloc[1:]
        else: # the case if two intervals are merged that have the same name => it should get back on stack
            new_rows = new_rows.iloc[0,:]
        stack = stack.append(new_rows)
        stack = stack.sort_values(by='start_time')
    return result

def _get_overlapping_activities(df, shift=1):
    """ gets all activities that have an overlap
    """
    assert shift >=1
    
    df = df.copy()
    df = df.sort_values(by=START_TIME)
    df = df.reset_index(drop=True)
    
    # get all activities that are have in an overlap
    mask = (df[START_TIME].shift(-shift) - df[END_TIME]) < pd.Timedelta('0ms')
    
    # as start_time is shifted upwards to select the right corresp. overlap 
    # shift the mask 'shift' steps downards  
    mask = mask.shift(+shift) | mask
    return df[mask] 

def correct_activities(df, strats=[], excepts=[]):
    """ gets df in form of activities and removes overlapping activities
    Parameters
    ----------
    df : pd.DataFrame
    strats : list
        the strategies for merging intervals with priority in ascending order
    excepts : list
        the activities which should be preserved when merging
    """
    df = df.copy()
    df = df.drop_duplicates(ignore_index=True)
    
    if _is_activity_overlapping(df):
        df, cor_lst = correct_activity_overlap(df, strats, excepts)
    else:
        cor_lst = None
    assert not _is_activity_overlapping(df)
    return df, cor_lst


START_TIME = 'start_time'
END_TIME = 'end_time'
TIME  = 'time'
NAME = 'name'
ACTIVITY = 'activity'
VAL = 'val'
DEVICE = 'device'

from pyadlml.dataset._datasets.fetch import (
    fetch_amsterdam,
    fetch_aras,
    fetch_casas_aruba,
    fetch_mitlab,
    fetch_tuebingen_2019,
    fetch_uci_adl_binary
)

from pyadlml.dataset._datasets.homeassistant import (
    load_homeassistant, load_homeassistant_devices
)

from pyadlml.dataset._datasets.activity_assistant import load as load_act_assist

from pyadlml.dataset.io import (
    set_data_home
)

from pyadlml.dataset._representations.changepoint import create_changepoint
from pyadlml.dataset._representations.raw import create_raw
from pyadlml.dataset._representations.image import (
    create_lagged_changepoint,
    create_lagged_lastfired,
    create_lagged_raw
)
from pyadlml.dataset._representations.lastfired import create_lastfired

from pyadlml.dataset import plot, plotly, io, obj, util


import pandas as pd
from pyadlml.dataset import TIME, DEVICE, VAL
import datetime

def load_homeassistant(db_url, limit=5000000, start_time=None, end_time=None):
    """ returns a dataframe representation of the homeassistant database
    Parameters
    ----------
    db_url : String
        filepath to homeassistant database
    limit : int
        the limit TODO don't know ^^
    start_time : str
        Datetime string to select the rows later that start_time
    Returns
    -------
    df : pd.DataFrame
    """
    if start_time == None and end_time == None:
        query = f"""
        SELECT entity_id, state, last_changed
        FROM states
        WHERE
            state NOT IN ('unknown', 'unavailable')
        ORDER BY last_changed DESC
        LIMIT {limit}
        """
    else:
        if end_time == None:
            end_time = datetime.datetime.now()
        query = f"""
        SELECT entity_id, state, last_changed
        FROM states
        WHERE
            state NOT IN ('unknown', 'unavailable') AND
            last_changed BETWEEN 
            '{start_time}' AND '{end_time}'
        ORDER BY last_changed ASC
        LIMIT {limit}        
        """
    df = pd.read_sql_query(query, db_url)
    return df

def load_homeassistant_devices(db_url, device_list, start_time=None, end_time=None):
    """ creates as dataframe in representation 1 from homeassistant database
    
    Parameters
    ----------
    db_url : String
        filepath to homeassistant database        
    device_list : lst
        device selection to filter
    start_time : pd.Timestamp
        the start time from when to filter
    end_time : pd.Timestamp
        the end time from when to filter
    Returns
    -------
    df : pd.DataFraem
    """
    
    df = load_homeassistant(db_url, start_time=start_time, end_time=end_time)
    df = df[df['entity_id'].isin(device_list)]
    df[TIME] = pd.to_datetime(df['last_changed'])
    df[VAL] = (df['state'] == 'on').astype(int)
    df[DEVICE] = df['entity_id']
    df = df[[TIME, DEVICE, VAL ]]
    return df


import unittest

from hassbrain_algorithm.controller import Controller
from hassbrain_algorithm.controller import Dataset
#from hassbrain_algorithm.algorithms.hmm import ProbabilityMassFunction
#from testing.testing import DiscreteHMM

from hassbrain_algorithm.datasets.homeassistant import DatasetHomeassistant


class TestHomeassistant(unittest.TestCase):

    def setUp(self):
        # set of observations
        self.ctrl = Controller()
        self.ctrl.set_dataset(Dataset.HASS)
        self.ctrl.load_dataset()
        self.hass_obj = self.ctrl._dataset #type: DatasetHomeassistant

#       algo = self.get_sel_algorithm()
#        dataset = self.get_dataset_by_name(dataset_name)
#        ctrl, dk, hmm_model = self._init_model_on_dataset(algo, dataset)
#        ctrl.set_dataset(dk)
#
#        if dataset_name == DATASET_NAME_HASS:
#            # get activities
#            act_list = []
#            for act in Activity.objects.all():
#                act_list.append(act.name)
#            dev_list = []
#            for dev in Device.objects.all():
#                if dev.location is not None:
#                    dev_list.append(dev.component.name + "." + dev.name)
#            print('*'*100)
#            print(act_list)
#            print(dev_list)
#            print('*'*100)
#            ctrl.set_custom_state_list(act_list)
#            ctrl.set_custom_obs_list(dev_list)
#
#        ctrl.load_dataset()
#        ctrl.register_model(hmm_model)

    def tearDown(self):
        pass

    def test_load_data(self):
        pass

    def test_print_hass_df(self):
        df = self.hass_obj._df
        print(DatasetHomeassistant.format_mat_full(df))

    def test_hashmaps(self):
        print(self.hass_obj.get_state_lbl_hashmap())
        print(self.hass_obj.get_obs_lbl_hashmap())

    def test_get_train_seqs(self):
        tr_seqs = self.hass_obj.get_train_seq()
        print(tr_seqs)
        print(self.hass_obj.decode_obs_seq(tr_seqs))

    def test_get_test_seq(self):
        test_seqs = self.hass_obj._test_seqs
        print(test_seqs)
        lbl_seqs, obs_seqs = self.hass_obj.get_test_labels_and_seq()
        print('-'*20)
        print('lbl_seqs: ', lbl_seqs)
        print('obs_seqs: ', obs_seqs)





import warnings
warnings.filterwarnings('ignore')
import matplotlib.pyplot as plt
plt.style.use('ggplot')
# Reference for customizing plots : http://matplotlib.org/users/customizing.html
# print(plt.style.available)

import numpy as np
from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import train_test_split


data = load_breast_cancer()
X = data.data
y = data.target
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size = .2)
print("Classes {}".format(np.unique(y)))


#-----------------------------------
feature_names = data.feature_names
X_train = X_train
X_test = X_test
#-----------------------------------
from sklearn.metrics import f1_score
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.linear_model import LogisticRegression

models = {'gb':GradientBoostingClassifier(),
          'mlp':MLPClassifier(),
          'knn':KNeighborsClassifier(),
          'reg':LogisticRegression()}

for model_key in models:
    model = models[model_key]
    model.fit(X_train, y_train)
    preds = model.predict(X_test)
    f1 = f1_score(y_test, preds)
    print("F1 for {0}: {1}".format(model_key, f1))


from skater.core.explanations import Interpretation
from skater.model import InMemoryModel

interpreter = Interpretation(X_test, feature_names=feature_names)
model = InMemoryModel(models['knn'].predict_proba, examples=X_train)
fig, ax = interpreter.feature_importance.save_plot_feature_importance(
    model,
    ascending=True)
fig.show()
#model = InMemoryModel(models['knn'].predict_proba, examples=X_train)
#fig, ax = interpreter.feature_importance.save_plot_feature_importance(
#    model,
#    ascending=True)
#fig.show()



import pandas as pd

from pyadlml.dataset.util import fill_nans_ny_inverting_first_occurence
from pyadlml.dataset.devices import correct_devices
from pyadlml.dataset.activities import correct_activities
                                        
from pyadlml.dataset.activities import correct_activity_overlap, \
    _is_activity_overlapping \

from pyadlml.dataset.obj import Data
from pyadlml.dataset import ACTIVITY, VAL, START_TIME, END_TIME, TIME, NAME, DEVICE

def _fix_data(path):
    """
    as the data is very inconsistent with tabs and spaces this is to make it alright again
    produces: 
        date time,id,value,activity 
    """
    data_path_tf = path[:-4] + 'corrected_data.csv'
    with open(path, 'r') as f_o, open(data_path_tf, 'w') as f_t:
        i= 1
        for line in f_o.readlines():
            s = line.split()                            
            # there is an error in line 1476694 where M014 is replaced with a 'c'
            if s[2] == 'c':
                s[2] = 'M014'
            # 'c' and '5' are randomly added onto values - remove them
            if 'c' in s[3]:
                s[3] = s[3].replace('c', '')                
            if '5' in s[3] and s[2][0] == 'M':
                s[3] = s[3].replace('5', '')
            if s[3] in ['ONM026', 'ONM009', 'ONM024']:         
                s[3] = 'ON'                
            # line 886912 error should be on
            if s[2][0] == 'M' and len(s[3]) == 1:
                s[3] = 'ON'                
            # line 900915 error should be off 
            if s[2][0] == 'M' and len(s[3]) == 2 and s[1] == '18:13:47.291404':
                s[3] = 'OFF'
            # add new line
            new_line = " ".join(s[:2]) + "," + ",".join(s[2:4])
            try:
                s[4] # test if there is an activity
                new_line += "," + " ".join(s[4:])
            except:
                pass
                
            f_t.write(new_line + "\n")
        f_t.close()
        f_o.close()


def _load_df(data_path):
    df = pd.read_csv(data_path,
                    sep=",",
                    #parse_dates=True,
                    infer_datetime_format=True,
                    na_values=True,
                    names=[START_TIME, 'id', VAL, ACTIVITY],
                    engine='python' #to ignore warning for fallback to python engine because skipfooter
                    #dtyp
                    )
    df[START_TIME] = pd.to_datetime(df[START_TIME])
    return df

def _val_activity_count(df_act):
    # confirm data assumptions from readme
    assert len(df_act[df_act['activity'] == 'Meal_Preparation']) == 1606
    # observed 2919 times line below
    len(df_act[df_act['activity'] == 'Relax'])# == 2910 # does not confirm 
    assert len(df_act[df_act['activity'] == 'Eating']) == 257
    assert len(df_act[df_act['activity'] == 'Work'])== 171
    assert len(df_act[df_act['activity'] == 'Sleeping']) == 401
    assert len(df_act[df_act['activity'] == 'Wash_Dishes']) == 65
    assert len(df_act[df_act['activity'] == 'Bed_to_Toilet']) == 157
    assert len(df_act[df_act['activity'] == 'Enter_Home']) == 431
    assert len(df_act[df_act['activity'] == 'Leave_Home']) == 431
    assert len(df_act[df_act['activity'] == 'Housekeeping']) == 33
    assert len(df_act[df_act['activity'] == 'Respirate']) == 6

def _get_devices_df(df):
    df_dev = df[(df['val'] == 'ON') | (df['val'] == 'OFF')]
    df_dev = df_dev.drop('activity', axis=1)
    df_dev['val'] = df_dev['val'] == 'ON'
    df_dev.columns = ['time', 'device', 'val']
    return df_dev

def load(data_path):
    df = _load_df(data_path)
    df_dev = _get_devices_df(df)
    df_act = _get_activity_df(df)
    
    dev_rep1 = correct_devices(df_dev)
    df_act, cor_lst = correct_activities(df_act)
    
    data = Data(df_act, dev_rep1)
    return data

def _get_activity_df(df):
    # get all rows containing activities
    df = df[~df[ACTIVITY].isnull()][[START_TIME, ACTIVITY]].copy()
    
    act_list = pd.unique(df[ACTIVITY])
    
    new_df_lst = []
    for i in range(1, len(act_list), 2):
        activity = act_list[i][:-4]
        act_begin = act_list[i-1]
        act_end = act_list[i]
           
        # create subsets for begin and end of chosen activity
        df_res = df[df[ACTIVITY] == act_begin].reset_index(drop=True)
        df_end = df[df[ACTIVITY] == act_end].reset_index(drop=True)
        
        # append sorted end_time to start_time as they should be
        # pairwise together
        df_res[ACTIVITY] = activity
        df_res[END_TIME] = df_end[START_TIME]
        new_df_lst.append(df_res)
    
    # data preparation
    res = pd.concat(new_df_lst)
    res = res.reindex(columns=[START_TIME, END_TIME, ACTIVITY])
    res = res.sort_values(START_TIME)
    res = res.reset_index(drop=True)
    return res


import os
import multiprocessing

ENV_PARALLEL='PYADLML_PARALLEL'

def set_parallel(val):
    """ tells code to run execute tasks with dask in parallel whenever possible
    """
    assert isinstance(val, bool)
    os.environ[ENV_PARALLEL] = str(val)
    assert get_parallel() == val

def get_parallel():
    val = os.environ.get(ENV_PARALLEL)
    assert val in ["True", "False"]
    return val == "True"
 	
def get_npartitions():
    """ returns the num of parallel threads that can work on a computer
    """
    return 4*multiprocessing.cpu_count()


import os
from pathlib import Path
import joblib
import hashlib
import zipfile
import shutil
from mega import Mega

DATA_DUMP_NAME = 'data.joblib'
ENV_DATA_HOME='PYADLML_DATA_HOME'

def set_data_home(path_to_folder):
    # TODO restrict to home folder and full paths
    os.environ[ENV_DATA_HOME] = path_to_folder

def get_data_home():
    return os.environ[ENV_DATA_HOME]

def load_from_data_home(param_dict):
    """
    Parameters
    ----------
    param_dict : dict
        contains values for X and y e.g frequency, name of dataset 
        example: 
            {'dataset': kasteren, 'freq':'10s', 'repr': 'raw'}
    Returns
    -------
    X : pd.DataFrame
    y : pd.DataFrame
    """
    # create folder name
    folder_name = hashdict2str(param_dict)
    folder_path = os.environ[ENV_DATA_HOME] + '/' + folder_name + '/'

    if not os.path.exists(folder_path):
        raise EnvironmentError # the dataset was not saved in this location

    # check if folder exists and return either result or raise an exception
    X = joblib.load(folder_path + 'X.joblib') 
    y = joblib.load(folder_path + 'y.joblib') 
   
    return X, y

def dump_in_data_home(X, y, param_dict):
    """
    Parameters
    ----------

    X : pd.DataFrame
    y : pd.DataFrame
    param_dict : dict
        contains values for X and y e.g frequency, name of dataset 
        example: 
            {'dataset': kasteren, 'freq':'10s', 'repr': 'raw'}
    """
    # create string representation of dictionary
    folder_name = hashdict2str(param_dict)
    folder_path = os.environ[ENV_DATA_HOME] + '/' + folder_name + '/'
    
    # remove folder if it already exists
    if os.path.exists(folder_path):        
        shutil.rmtree(folder_path)
        
    _create_folder(folder_path)

    # save all data
    joblib.dump(X, folder_path + 'X.joblib') 
    joblib.dump(y, folder_path + 'y.joblib') 
    joblib.dump(param_dict, folder_path + 'param_dict.joblib') 


def _create_folder(path_to_folder):
    Path(path_to_folder).mkdir(parents=True, exist_ok=True)

def hashdict2str(param_dict):
    """ creates a unique string for a dictionary 
    Parameters
    ----------
    param_dict : dict
        Contains the attributes of the encoder and the data representations
    Returns
    -------
    folder_name : str
    """
    # sort dictionary after keys for determinism
    param_dict = {k: v for k, v in sorted(param_dict.items(), key=lambda item: item[1])}

    # create string
    param_string = str(param_dict).encode('utf-8')
    folder_name = hashlib.md5(param_string).hexdigest()
    return str(folder_name)

def _delete_data(path_to_folder):
    # make shure only data in home directorys are deleted
    assert '/home/' == path_to_folder[:6] 
    shutil.rmtree(path_to_folder)

def _data_2_folder_name(path_to_folder, data_name):
    param_dict = {'dataset' : data_name}
    folder_name = hashdict2str(param_dict)
    folder_path = path_to_folder + '/' + folder_name
    return folder_path

def fetch_handler(keep_original, cache, dataset_name, 
        mega_filename, mega_url,
        load_func, data_postfix=''):
    """ handles the downloading, loading and caching of a dataset
    Parameters
    ----------
    Returns
    -------
    data : object

    """
    data_home = get_data_home()
    data_home_dataset = ''.join([data_home, '/', dataset_name])
    cache_data_folder = _data_2_folder_name(data_home, dataset_name)

    # download data    
    if not os.path.isdir(data_home_dataset):
        # download file from mega # TODO make official way available
        _download_from_mega(get_data_home(), mega_filename, mega_url)

    
    # load data
    if data_postfix != '':
        data_name = cache_data_folder + '/' \
            + DATA_DUMP_NAME[:-7] + '_' + data_postfix + '.joblib'
    else:
        data_name = cache_data_folder + '/' + DATA_DUMP_NAME

    if Path(data_name).is_file(): 
        data = joblib.load(data_name) 
    else:
        data = load_func(data_home_dataset + '/')
        if cache:
            _create_folder(cache_data_folder)
            joblib.dump(data, data_name)
            Path(cache_data_folder + '/' + dataset_name).touch()


    # clean up data
    # TODO note that the folder is deleted. For two different subjects
    # caching one and deleting another leads to deletion of the first 
    if not cache and os.path.exists(cache_data_folder):
        _delete_data(cache_data_folder)
    if not keep_original and os.path.exists(data_home_dataset):
        _delete_data(data_home_dataset)

    return data





def clear_data_home():
    """ Delete all the content of the data home cache.
    """
    data_home = get_data_home()
    _delete_data(data_home) 
    Path.mkdir(data_home)

def _download_from_mega(data_home, file_name, url):
    """ downloads dataset from MEGA and extracts it
    """
    file_dp = data_home + '/' + file_name
    
    # download from mega
    m = Mega()    
    m.download_url(url, dest_path=data_home, dest_filename=file_name)

    # unzip data 
    with zipfile.ZipFile(file_dp,"r") as zip_ref:
        zip_ref.extractall(data_home)

    # remove zip file
    Path(file_dp).unlink()


import pandas as pd
import numpy as np
"""
    includes generic methods for manpulating dataframes
"""

def print_df(df):
    with pd.option_context('display.max_rows', None, 'display.max_columns', None):  # more options can be specified also
        print(df)


def timestr_2_timedeltas(t_strs):
    """
        gets either a string or a list of strings to convert to a list of 
        timedeltas
    """
    if isinstance(t_strs, list):
        return [timestr_2_timedelta(t_str) for t_str in t_strs]
    else:
        return [timestr_2_timedelta(t_strs)]


def timestr_2_timedelta(t_str):
    """
        t_str (string)
        of form 30s, 30m
    """
    ttype = t_str[-1:]
    val = int(t_str[:-1])

    assert ttype in ['h','m','s']
    assert (val > 0 and val <=12 and ttype == 'h')\
        or (val > 0 and val <= 60 and ttype == 'm')\
        or (val > 0 and val <= 60 and ttype == 's')
    import datetime as dt
    if ttype == 's':
        return pd.Timedelta(seconds=val)
    if ttype == 'm':
        return pd.Timedelta(seconds=val*60)
    if ttype == 'h':
        return pd.Timedelta(seconds=val*3600)


def time2int(ts, t_res='30m'):
    """
    rounds to the next lower min bin or hour bin
    """
    assert t_res[-1:] in ['h','m']
    val = int(t_res[:-1])

    assert (val > 0 and val <=12 and t_res[-1:] == 'h')\
        or (val > 0 and val <= 60 and t_res[-1:] == 'm')

    import datetime as dt
    zero = dt.time()

    if t_res[-1:] == 'h':
        hs = val
        h_bin = int(ts.hour/hs)*hs
        return dt.time(hour=h_bin)

    elif t_res[-1:] == 'm':
        ms = val
        m_bin = int(ts.minute/ms)*ms
        return dt.time(hour=ts.hour, minute=m_bin)
    else:
        raise ValueError



def fill_nans_ny_inverting_first_occurence(df):
    """
    fills up true or false values
    :param df:
                Name           0      1         10
        time                                     
        2008-02-25 00:20:14    NaN    NaN   ...    True
        2008-02-25 00:22:57    NaN    NaN   ...    False
        2008-02-25 09:33:41    NaN    True  ...    True
        2008-02-25 09:33:42    False    NaN   ...    False
    :return:
                Name           0      1        10   
        Time                                       
        2008-02-25 00:20:14    True   False ... True 
        2008-02-25 00:22:57    True   False ... False 
        2008-02-25 09:33:41    True   True  ... True 
        2008-02-25 09:33:42    False   NaN  ... False 
    """
    for col_label in df.columns:
        col = df[col_label]

        # get timestamp of first valid index and replace previous Nans 
        #   by opposite
        ts = col.first_valid_index()
        idx = df.index.get_loc(ts)
        col.iloc[0:idx] = not col[ts]
    return df





from hassbrain_algorithm.datasets._dataset import DataRep, _Dataset
from hassbrain_algorithm.controller import Controller
from hassbrain_algorithm.controller import Dataset
from hassbrain_algorithm.datasets.homeassistant.hass import DatasetHomeassistant


SEC1 ='0.01666min'
SEC2 ='0.03333min'
SEC3 ='0.05min'
SEC6 = '0.1min'
SEC12 = '0.2min'
SEC30 = '0.5min'
SEC20 = '0.3min'
SEC45 = '0.75min'
MIN1 = '1min'
MIN30 = '30min'

TEST_SEL_ALL = 'all'
TEST_SEL_ODO = 'one_day_out'

BASE_PATH = '/home/cmeier/code/data/thesis_results'

# starts at 2019-06-29
# ends at 2019-07-11
DN_HASS_CHRIS = 'hass_chris'

# starts at 2019-07-12T04:31:28
# ends at 2019-07-23T04:31:28
DN_HASS_CHRIS_FINAL = 'hass_chris_final'
DK_HASS = Dataset.HASS
TEST_DAY_HASS_CHRIS_FINAL_0 = '2019-07-10'
TEST_DAY_HASS_CHRIS_FINAL_1 = '2019-07-08' # 27 activities


DT_RAW = DataRep.RAW
DT_CH = DataRep.CHANGEPOINT
DT_LF = DataRep.LAST_FIRED

#------------------------
# Configuration


DATA_NAME = DN_HASS_CHRIS_FINAL
DK = DK_HASS
DATA_TYPE = DT_CH
TEST_SEL = TEST_SEL_ODO
TIME_DIFF = SEC20
TEST_DAY = TEST_DAY_HASS_CHRIS_FINAL_1
#------------------------


DATASET_FILE_NAME = 'dataset_' + DATA_NAME + '_' + DATA_TYPE.value + '_' + TIME_DIFF + ".joblib"
DATASET_FILE_PATH = BASE_PATH + '/' + DATA_NAME + '/datasets/' + DATASET_FILE_NAME

#DATASET_FILE_PATH = '/home/cmeier/code/hassbrain_algorithm/testing/test_files/' + DATASET_FILE_NAME

def join_two_seperate_databses(dataset):
    dataset._acts.load_basic()
    dataset._devs.load_basic()
    dataset._devs._sens_file_path = '/home/cmeier/code/data/hassbrain/datasets/hass_chris_final/data/home-assistant_v2_pre.db'

    df1, df1_hm, df1_rev_hm = dataset._devs._load_basic()
    df2 = dataset._devs._dev_data
    df2_rev_hm = dataset._devs._sensor_label_reverse_hashmap

    res_df = dataset._devs.concatenate(
        df1.copy(),
        df2.copy(),
        df1_rev_hm,
        df2_rev_hm,
        df1_hm
    )
    dataset._devs._dev_data = res_df

def load_rest(dataset):
    test_all_x, test_all_y = dataset._load_all_labeled_data()
    dataset._test_all_x = test_all_x
    dataset._test_all_y = test_all_y

    train_y = None
    test_x = None
    test_y = None

    if dataset._test_sel == 'all':
        dataset._devs.train_eq_test_dat()
        dataset._acts.train_eq_test_dat()
    elif dataset._test_sel == 'one_day_out':
        if dataset._test_day is None:
            test_day = dataset._acts.get_random_day()
        else:
            import datetime
            test_day = datetime.datetime.strptime(
                dataset._test_day, '%Y-%m-%d').date()

        dataset._devs.split_train_test_dat(test_day)
        dataset._acts.split_train_test_dat(test_day)
    else:
        raise ValueError

    train_y, test_x, test_y = dataset._get_specific_repr()
    assert train_y is not None
    assert test_x is not None
    assert test_y is not None

    dataset._train_y = train_y
    dataset._test_x = test_x
    dataset._test_y = test_y



def main():
    dk = DK
    data_name = DATA_NAME

    # set of observations
    ctrl_config = {
        'datasets': {
            'kasteren': {
                'path_to_config': '/home/cmeier/code/data/hassbrain/datasets/kasteren/config.yaml'
            },
            'hass_testing': {
                'path_to_config': '/home/cmeier/code/data/hassbrain/datasets/hass_testing/config.yaml'
            },
            'hass_testing2': {
                'path_to_config': '/home/cmeier/code/data/hassbrain/datasets/hass_testing2/config.yaml'
            },
            'hass_chris': {
                'path_to_config': '/home/cmeier/code/data/hassbrain/datasets/hass_chris/config.yaml'
            },
            'hass_chris_final': {
                'path_to_config': '/home/cmeier/code/data/hassbrain/datasets/hass_chris_final/config.yaml'
            }
        }
    }
    ctrl = Controller(config=ctrl_config)
    params = {'repr': DATA_TYPE,
              'data_format': 'bernoulli',
              'test_selection': TEST_SEL,
              'test_day': TEST_DAY,
              'freq': TIME_DIFF,
              'include_idle':False}
    ctrl.set_dataset(
        data_name=data_name,
        data_type=dk,
        params=params
    )
    dataset = ctrl._dataset # type: _Dataset
    join_two_seperate_databses(dataset)
    load_rest(dataset)
    #act_stats = dataset.get_act_stats()
    #dev_stats = dataset.get_dev_stats()
    ctrl.save_dataset(DATASET_FILE_PATH)
    #print('filepath: ', DATASET_FILE_PATH)
    print('*'*10)






if __name__ == '__main__':
    main()
    print('finished')


from pyadlml.dataset._representations.raw import create_raw
from pyadlml.dataset._representations.changepoint import create_changepoint
from pyadlml.dataset.activities import check_activities

class Data():
    def __init__(self, activities, devices):
        #assert check_activities(activities) 
        #assert check_devices(devices)

        self.df_activities = activities
        self.df_devices = devices

        # list of activities and devices
        #self.activities = list(activities.activity.unique())
        #self.devices = list(devices.device.unique())
        self.df_raw = None
        self.df_cp = None
        self.df_lf = None

    def create_cp(self, t_res):
        raise NotImplementedError

    def create_raw(self, t_res=None, idle=False):
        self.df_raw = create_raw(self.df_devices, self.df_activities, t_res)

    def create_lastfired(self):
        raise NotImplementedError


import plotly.figure_factory as ff
import plotly.express as px
import numpy as np
import plotly.express as px
import plotly.graph_objects as go
from plotly.colors import n_colors

from pyadlml.dataset import START_TIME
from pyadlml.dataset.stats.activities import activities_duration_dist, \
    activities_count, activities_durations, activities_dist, \
    activities_transitions

def ridge_line(df_act, t_range='day', n=1000):
    """
    https://plotly.com/python/violin/

    for one day plot the activity distribution over the day
    - sample uniform from each interval   
    """
    df = activities_dist(df_act.copy(), t_range, n)

    colors = n_colors('rgb(5, 200, 200)', 'rgb(200, 10, 10)', len(df.columns), colortype='rgb')
    data = df.values.T

    fig = go.Figure()
    i = 0
    for data_line, color in zip(data, colors):
        fig.add_trace(go.Violin(x=data_line, line_color=color, name=df.columns[i]))
        i += 1

    fig.update_traces(orientation='h', side='positive', width=3, points=False)
    fig.update_layout(xaxis_showgrid=False, xaxis_zeroline=False)
    return fig

def boxplot_duration(df_act, y_scale='norm'):
    """
        plot a boxplot of activity durations (mean) max min 
    """
    assert y_scale in ['norm', 'log']

    df = activities_duration_dist(df_act)

    # add column for display of datapoints later
    df[START_TIME] = df_act[START_TIME].dt.strftime('%c')

    if y_scale == 'log':
        df['log minutes'] = np.log(df['minutes'])
        labels={'minutes': 'log minutes'}
    else:
        labels={'minutes': 'minutes'}    

    fig = px.box(df, x="activity", y=labels['minutes'], 
                 notched=True, # used notched shape
                 labels=labels,
                 points='all', # display points next to box plot
                 title="Activity durations",
                 hover_data=[START_TIME, 'minutes'] # add day column to hover data
                )
    return fig

def hist_cum_duration(df_act, y_scale='both'):
    """
    plots the cummulated activities durations in a histogram for each activity 

    """
    assert y_scale in ['norm', 'log', 'both']

    title = 'Activity cummulative durations'

    act_dur = activities_durations(df_act.copy())
    df = act_dur[['minutes']]
    df.reset_index(level=0, inplace=True)
    if y_scale in ['norm', 'log']: 
        if y_scale == 'log':
            df['minutes'] = np.log(df['minutes'])
            labels={'minutes': 'log minutes'}
        else:
            labels={'minutes': 'minutes'}
            
        df = df.sort_values(by=['minutes'], axis=0)
        fig = px.bar(df, y='activity', x='minutes', 
                    title=title,
                    labels=labels, 
                    height=400,
                    #hover_data=['fraction'] TODO add the fraction by hovering
                    )
    else:
        df = df.sort_values(by=['minutes'], axis=0)
        col_label = 'minutes'
        col_label_log = 'log minutes'
        col_activity = 'activity'
        df[col_label_log] = np.log(df[col_label])
        
        fig = go.Figure()
        fig.add_trace(go.Bar(x=df[col_label], y=df[col_activity], \
                                orientation='h', name='duration'))
        fig.add_trace(go.Bar(x=df[col_label_log], y=df[col_activity], \
                                orientation='h', name='log duration', \
                                visible=False))
        # Add dropdown
        fig.update_layout(
            title=title,
            xaxis_title="duration",
            yaxis_title='activities',
            updatemenus=[
                dict(
                    buttons=list([
                        dict(
                            label = "no scaling", method = "update",
                            args = [{"visible": [True, False]},
                                {"title": title, "xaxis_title": 'asdf'}]),
                        dict(
                            label = "log scaled", method = "update", 
                            args = [{"visible": [False, True]},
                                {"title": title, "xaxis_title": 'asdf'}])
                    ]),
                    active=0,
                    direction="up",
                    showactive=True,
                    x=1.0,
                    xanchor="right",
                    y=-0.2,
                    yanchor="bottom"
                ),
            ]
        )

    return fig

def hist_counts(df_act, y_scale='norm'):
    """
    plots the activities durations against each other
    """
    assert y_scale in ['norm', 'log']

    col_label = 'occurence'
    title ='activity occurrences'

    df = activities_count(df_act.copy())
    df.reset_index(level=0, inplace=True)
    
    if y_scale == 'log':
        df[col_label] = np.log(df[col_label])
        labels={col_label: 'log count'}
    else:
        labels={col_label: 'count'}    

    df = df.sort_values(by=['occurence'], axis=0)
    fig = px.bar(df, title=title, y='activity', x=col_label, orientation='h', labels=labels, height=400)
    return fig

def gantt(df):
    """
    """
    df_plt = df.copy()
    df_plt.columns = ['Start', 'Finish', 'Resource']
    df_plt['Task'] = 'Activity'
    fig = ff.create_gantt(df_plt, 
        index_col='Resource',
        show_colorbar=True, 
        bar_width=0.2, 
        height=500,
        width=None,
        group_tasks=True,
        showgrid_x=True, showgrid_y=False)

    # Add range slider
    btn_12h = dict(count=12, label="12h", step="hour", stepmode="backward")
    btn_1d = dict(count=1, label="1d", step="day", stepmode="backward")
    btn_1w = dict(count=7, label="1w", step="day", stepmode="backward")
    btn_1m = dict(count=1, label="1m", step="month", stepmode="backward")
    #btn_all =
    fig.update_layout(
        xaxis=dict(
            rangeselector=dict(
                buttons=list([btn_12h, btn_1d, btn_1w, btn_1m])#, dict(step="all")
            ),
            rangeslider=dict(
                visible=True
            ),
            type="date"
        )
    )    
    return fig

def heatmap_transitions(df_act, z_scale='norm'):
    """
    """
    assert z_scale in ['norm', 'log'], 'z-scale has to be either norm or log'

    name = 'Activity transitions'

    # get the list of cross tabulations per t_window
    df = activities_transitions(df_act)
    act_lst = list(df.columns)
    if z_scale == 'norm':
        z = df.values
    elif z_scale == 'log':
        z = np.log(df.values)
    else:
        raise ValueError

    fig = go.Figure(data=go.Heatmap(
        #labels=dict(x=x_label, y=y_label, color=color),
        z=z,
        x=act_lst,
        y=act_lst,
        colorscale='Viridis',
        hoverongaps = False))
    fig.update_layout(title=name)
    return fig


import numpy as np
import pandas as pd

from pyadlml.dataset.util import fill_nans_ny_inverting_first_occurence
from pyadlml.dataset.obj import Data

from pyadlml.dataset.activities import correct_activity_overlap,\
    _is_activity_overlapping

from pyadlml.dataset import  ACTIVITY, VAL, \
    START_TIME, END_TIME, TIME, NAME, DEVICE

from pyadlml.dataset.activities import correct_activities
from pyadlml.dataset.devices import correct_devices



def _load_activities(activity_fp):
    act_map = {
        1: 'leave house',
        4: 'use toilet',
        5: 'take shower',
        10:'go to bed',
        13:'prepare Breakfast',
        15:'prepare Dinner',
        17:'get drink'
        }
    ide = 'id'
    df = pd.read_csv(activity_fp,
                     sep="\t",
                     skiprows=23,
                     skipfooter=1,
                     parse_dates=True,
                     names=[START_TIME, END_TIME, ide],
                     engine='python') 
    df[START_TIME] = pd.to_datetime(df[START_TIME])
    df[END_TIME] = pd.to_datetime(df[END_TIME])
    df[ACTIVITY] = df[ide].map(act_map)
    df = df.drop(ide, axis=1)
    return df


def _load_devices(device_fp):
    sens_labels = {
        1: 'Microwave', 
        5: 'Hall-Toilet door',
        6: 'Hall-Bathroom door',
        7: 'Cups cupboard',
        8: 'Fridge',
        9: 'Plates cupboard',
        12: 'Frontdoor',
        13: 'Dishwasher',
        14: 'ToiletFlush',
        17: 'Freezer',
        18: 'Pans Cupboard',
        20: 'Washingmachine',
        23: 'Groceries Cupboard',
        24: 'Hall-Bedroom door'
    }
    ide = 'id'
    sens_data = pd.read_csv(device_fp,
                sep="\t",
                skiprows=23,
                skipfooter=1,
                parse_dates=True,
                names=[START_TIME, END_TIME, ide, VAL],
                engine='python' #to ignore warning for fallback to python engine because skipfooter
                #dtype=[]
                )


    #k todo declare at initialization of dataframe
    sens_data[START_TIME] = pd.to_datetime(sens_data[START_TIME])
    sens_data[END_TIME] = pd.to_datetime(sens_data[END_TIME])
    sens_data[VAL] = sens_data[VAL].astype('bool')
    sens_data = sens_data.drop(VAL, axis=1)

    # replace numbers with the labels
    sens_data[DEVICE] = sens_data[ide].map(sens_labels)
    sens_data = sens_data.drop(ide, axis=1)
    sens_data = sens_data.sort_values(START_TIME)
    return sens_data


def load(device_fp, activity_fp):
    df_act = _load_activities(activity_fp)
    df_dev = _load_devices(device_fp)
    df_act, cor_lst = correct_activities(df_act)
    
    df_dev = correct_devices(df_dev)

    data = Data(df_act, df_dev)
    data.correction_activities = cor_lst
    
    return data




import plotly.figure_factory as ff
import plotly.express as px
import numpy as np
from pyadlml.dataset.stat import devices_count
"""
#1.  categorical dot plot for device activations over time
        - color of dot could be the activity
        - next to activity bar plot

2. dot plot activities vs labeled 

3. confusion matrix as heatmap
"""


def gantt_activities_vs_devices(df):
    """
    """
    df_plt = df.copy()
    df_plt.columns = ['Start', 'Finish', 'Task']
    df_plt['Resource'] = 'A'
    fig = ff.create_gantt(df_plt, 
    colors=['#333F44', '#93e4c1'],
    index_col='Resource',
    show_colorbar=True, 
    bar_width=0.2, 
    group_tasks=True,
    showgrid_x=True, showgrid_y=True)
    return fig


import pandas as pd

# ------------ Configuration
from hassbrain_algorithm.controller import Dataset


BHMM = 'bhmm'
BHMMPC = 'bhmmpc'
BHSMM = 'bhsmm'
BHSMMPC = 'bhsmmpc'
HMM = 'hmm'
HMMPC = 'hmmpc'
HSMM = 'hsmm'
HSMMPC = 'hsmmpc'
TADS = 'tads'

SEC1 ='0.01666min'
SEC6 = '0.1min'
SEC30 = '0.5min'
MIN1 = '1min'

DK = Dataset.HASS

DN_HASS_CHRIS = 'hass_chris'
DN_HASS_CHRIS_FINAL = 'hass_chris_final'
BASE_PATH = '/home/cmeier/code/data/thesis_results'

DT_RAW = 'raw'
DT_CH = 'changed'
DT_LF = 'last_fired'

#------------------------
# Configuration

MODEL_CLASS = BHMM
DATA_NAME = DN_HASS_CHRIS_FINAL
DATA_TYPE = DT_RAW
TIME_DIFF = SEC30
BEST_MODEL_SELECTION = True

# ------------------------------
MODEL_NAME = 'model_' + MODEL_CLASS + '_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF
if BEST_MODEL_SELECTION:
    MODEL_FOLDER_PATH = BASE_PATH + '/' + DATA_NAME + '/best_model_selection/' + MODEL_NAME
else:
    MODEL_FOLDER_PATH = BASE_PATH + '/' + DATA_NAME + '/models/' + MODEL_NAME

CONF_MAT_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME +".confusion_matrix.numpy"

df = pd.read_csv(CONF_MAT_FILE_PATH)

cols = df.columns

df.set_index('Unnamed: 0')

new_cols = ['ConfMat']
for col in cols[1:]:
    words = col.split('_')
    letters = ''
    for word in words:
        letters = letters + word[0]
    new_cols.append(letters)

df.columns = new_cols

texcode = df.to_latex(index=False)

tex_file_name = CONF_MAT_FILE_PATH + '.tex'
f = open(tex_file_name, "w")
f.write(texcode)
f.close()


import numpy as np
import pandas as pd

class LeaveOneDayOut():
    def __init__(self, num_days=1):
        """
        
        """
        self.n_splits = num_days
    
    def get_n_splits(self, X=None, y=None, groups=None):
        """Returns the number of splitting iterations in the cross-validator
        Parameters
        ----------
        X : object
            Always ignored, exists for compatibility.
        y : object
            Always ignored, exists for compatibility.
        groups : array-like of shape (n_samples,)
            Group labels for the samples used while splitting the dataset into
            train/test set. This 'groups' parameter must always be specified to
            calculate the number of splits, though the other parameters can be
            omitted.
        Returns
        -------
        n_splits : int
            Returns the number of splitting iterations in the cross-validator.
        """
        return self.n_splits
    
    def split(self, X=None, y=None, groups=None):
        """ Generate indices to split data into training and test set.
        Parameters
        ----------
        X : pd.DataFrame
            with timeindex 
        y : pd.Series
            with timeindex 
        Returns 
        -------
        splits : list
            Returns tuples of splits of train and test sets
            example: [(train1, test1), ..., (trainn, testn)]
        """
        
        # get all days
        df = X.copy()
        df = df.sort_values(by='time')
        df = df.reset_index()['time'].dt.floor('d').value_counts().rename_axis('date')
        days = df.reset_index(name='count')['date']
        
        df = X.copy()
        df = df.reset_index()

        res = []
        for i in range(self.n_splits):
            # select uniformly a random day
            rnd_idx = np.random.randint(0, high=len(days)-1)
            rnd_day = days.iloc[rnd_idx]
            
            # get indicies of all data for that day and the others
            rnd_dayp1 = rnd_day + pd.Timedelta('1D')
            mask = (rnd_day < df['time']) & (df['time'] < rnd_dayp1)
            idxs_test = df[mask].index.values
            idxs_train = df[~mask].index.values
            
            res.append((idxs_train, idxs_test))
        
        return res


from hassbrain_algorithm.controller import Controller, Dataset
from hassbrain_algorithm.datasets._dataset import DataRep
from hassbrain_algorithm.models.hmm.bhmm import BernoulliHMM

# model names
from scripts.test_model import BHMMTestModel

BHMM = 'bhmm'
BHMMPC = 'bhmm_pc'
BHSMM = 'bhsmm'
BHSMMPC = 'bhsmmpc'
HMM = 'hmm'
HMMPC = 'hmmpc'
HSMM = 'hsmm'
HSMMPC = 'hsmmpc'

SEC1 ='0.01666min'
SEC2 ='0.03333min'
SEC3 ='0.05min'
SEC6 = '0.1min'
SEC12 = '0.2min'
SEC30 = '0.5min'

MIN1 = '1min'
MIN30 = '30min'
DK = Dataset.HASS

DN_KASTEREN = 'kasteren'
DN_HASS_TESTING2 = 'hass_testing2'
DN_HASS_CHRIS = 'hass_chris'
DN_HASS_CHRIS_FINAL = 'hass_chris_final'
BASE_PATH = '/home/cmeier/code/data/thesis_results'

DT_RAW = 'raw'
DT_CH = 'changed'
DT_LF = 'last_fired'

#------------------------
# Configuration

MODEL_CLASS = BHMMPC
DATA_NAME = DN_HASS_CHRIS_FINAL
DATA_TYPE = DT_CH
TIME_DIFF = SEC6

#------------------------

DATASET_FILE_NAME = 'dataset_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF + ".joblib"
DATASET_FILE_PATH = BASE_PATH + '/' + DATA_NAME + '/datasets/' + DATASET_FILE_NAME

MODEL_NAME = 'model_' + MODEL_CLASS + '_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF
# todo this
#MODEL_FOLDER_PATH = BASE_PATH + '/' + DATA_NAME + '/models/' + MODEL_NAME
MODEL_FOLDER_PATH = BASE_PATH + '/' + DATA_NAME + '/best_model_selection/' + MODEL_NAME

MODEL_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME +".joblib"
MD_LOSS_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.loss.csv'
MD_LOSS_IMG_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.loss.png'
MD_INFST_IMG_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.inferred_states.png'
MD_CONF_MAT_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.confusion_matrix.numpy'
MD_METRICS_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.metrics.csv'
MD_ACT_DUR_DISTS_IMG_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.act_dur_dists.png'
MD_ACT_DUR_DISTS_DF_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.act_dur_dists.csv'
DATA_ACT_DUR_DISTS_IMG_FILE_PATH = MODEL_FOLDER_PATH + '/dataset.act_dur_dists.png'
DATA_ACT_DUR_DISTS_DF_FILE_PATH = MODEL_FOLDER_PATH + '/dataset.act_dur_dists.csv'

def load_domain_knowledge(file_path):
    with open(file_path, 'r') as file:
        import json
        data = json.load(file)
        act_data = data['activity_data']
        loc_data = data['loc_data']

    import datetime
    for act_data_point in act_data:
        act_data_point['start'] = datetime.datetime.strptime(
            act_data_point['start'],
            '%H:%M:%S').time()
        act_data_point['end'] = datetime.datetime.strptime(
            act_data_point['end'],
            '%H:%M:%S').time()

    return act_data, loc_data

def main():
    ctrl = Controller()
    ctrl.load_dataset_from_file(DATASET_FILE_PATH)

    #hmm_model = BHMMTestModel(ctrl)
    #hmm_model = BHSMMTestModel(ctrl)
    from hassbrain_algorithm.models.hmm.bhmm_hp import BernoulliHMM_HandcraftedPriors
    hmm_model = BernoulliHMM_HandcraftedPriors(ctrl)

    ctrl.register_model(hmm_model, MODEL_NAME)

    # load domain knowledge
    path = '/home/cmeier/code/data/hassbrain/datasets/hass_chris_final/data/domain_knowledge.json'
    act_data, loc_data = load_domain_knowledge(path)
    ctrl.register_location_info(MODEL_NAME, loc_data)
    ctrl.register_activity_info(MODEL_NAME, act_data)

    # load model
    #ctrl.load_model(MODEL_FILE_PATH, MODEL_NAME)
    from scripts.test_model import BHSMMTestModel


    ctrl.register_benchmark(MODEL_NAME)

    ctrl.init_model_on_dataset(MODEL_NAME)
    ctrl.register_loss_file_path(MD_LOSS_FILE_PATH, MODEL_NAME)
    ctrl.train_model(MODEL_NAME)

    # bench the model
    reports = ctrl.bench_models()

    # save metrics
    ctrl.save_df_metrics_to_file(MODEL_NAME, MD_METRICS_FILE_PATH)
    ctrl.save_df_confusion(MODEL_NAME, MD_CONF_MAT_FILE_PATH)
    ctrl.save_df_act_dur_dists(MODEL_NAME, MD_ACT_DUR_DISTS_DF_FILE_PATH,
                               DATA_ACT_DUR_DISTS_DF_FILE_PATH)

    # plots
    ctrl.save_plot_trainloss(MD_LOSS_IMG_FILE_PATH, MODEL_NAME)
    ctrl.plot_and_save_inferred_states(MD_INFST_IMG_FILE_PATH, MODEL_NAME)
    ctrl.save_plot_act_dur_dists(MODEL_NAME, MD_ACT_DUR_DISTS_IMG_FILE_PATH,
                                 DATA_ACT_DUR_DISTS_IMG_FILE_PATH)

if __name__ == '__main__':
    print('dataset filepath : ', DATASET_FILE_PATH)
    print('model filepath: ', MODEL_FILE_PATH)
    main()
    print('finished')


from hassbrain_algorithm.datasets._dataset import DataRep
from hassbrain_algorithm.controller import Controller
from hassbrain_algorithm.controller import Dataset
from hassbrain_algorithm.datasets.homeassistant.hass import DatasetHomeassistant


SEC1 ='0.01666min'
SEC2 ='0.03333min'
SEC3 ='0.05min'
SEC6 = '0.1min'
SEC12 = '0.2min'
SEC30 = '0.5min'
SEC45 = '0.75min'
MIN1 = '1min'
MIN30 = '30min'

TEST_SEL_ALL = 'all'
TEST_SEL_ODO = 'one_day_out'

BASE_PATH = '/home/cmeier/code/data/thesis_results'
DN_KASTEREN = 'kasteren'
DN_HASS_TESTING2 = 'hass_testing2'
DN_HASS_CHRIS = 'hass_chris'

DK_HASS = Dataset.HASS
DK_KASTEREN = Dataset.KASTEREN
TEST_DAY_KASTEREN_0 = '2008-03-15'
TEST_DAY_HA = '2008-03-15'

DT_RAW = DataRep.RAW
DT_CH = DataRep.CHANGEPOINT
DT_LF = DataRep.LAST_FIRED

#------------------------
# Configuration


DATA_NAME = DN_HASS_CHRIS
DK = DK_HASS
DATA_TYPE = DT_RAW
TEST_SEL = TEST_SEL_ODO
TIME_DIFF = MIN30
TEST_DAY = None
#------------------------


DATASET_FILE_NAME = 'dataset_' + DATA_NAME + '_' + DATA_TYPE.value + '_' + TIME_DIFF + ".joblib"
DATASET_FILE_PATH = BASE_PATH + '/' + DATA_NAME + '/datasets/' + DATASET_FILE_NAME

#DATASET_FILE_PATH = '/home/cmeier/code/hassbrain_algorithm/testing/test_files/' + DATASET_FILE_NAME

def main():
    dk = DK
    data_name = DATA_NAME

    # set of observations
    ctrl_config = {
        'datasets': {
            'kasteren': {
                'path_to_config': '/home/cmeier/code/data/hassbrain/datasets/kasteren/config.yaml'
            },
            'hass_testing': {
                'path_to_config': '/home/cmeier/code/data/hassbrain/datasets/hass_testing/config.yaml'
            },
            'hass_testing2': {
                'path_to_config': '/home/cmeier/code/data/hassbrain/datasets/hass_testing2/config.yaml'
            },
            'hass_chris': {
                'path_to_config': '/home/cmeier/code/data/hassbrain/datasets/hass_chris/config.yaml'
            }
        }
    }
    ctrl = Controller(config=ctrl_config)
    params = {'repr': DATA_TYPE,
              'data_format': 'bernoulli',
              'test_selection': TEST_SEL,
              'test_day': TEST_DAY,
              'freq': TIME_DIFF}
    ctrl.set_dataset(
        data_name=data_name,
        data_type=dk,
        params=params
    )
    ctrl.load_dataset()
    tmp1 = ctrl._dataset.get_dev_stats()
    tmp2 = ctrl._dataset.get_act_stats()
    #ctrl.save_dataset(DATASET_FILE_PATH)
    #print('filepath: ', DATASET_FILE_PATH)
    print('*'*10)

if __name__ == '__main__':
    main()
    print('finished')


from hassbrain_algorithm.controller import Controller, Dataset
from hassbrain_algorithm.datasets._dataset import DataRep, _Dataset
from hassbrain_algorithm.models.hmm.bhmm import BernoulliHMM

# model names
from scripts.test_model import BHMMTestModel

BHMM = 'bhmm'
BHMMPC = 'bhmm_pc'
BHSMM = 'bhsmm'
BHSMMPC = 'bhsmmpc'
HMM = 'hmm'
HMMPC = 'hmmpc'
HSMM = 'hsmm'
HSMMPC = 'hsmmpc'

SEC1 ='0.01666min'
SEC2 ='0.03333min'
SEC3 ='0.05min'
SEC6 = '0.1min'
SEC12 = '0.2min'
SEC30 = '0.5min'

MIN1 = '1min'
MIN30 = '30min'
DK = Dataset.HASS

DN_KASTEREN = 'kasteren'
DN_HASS_TESTING2 = 'hass_testing2'
DN_HASS_CHRIS = 'hass_chris'
DN_HASS_CHRIS_FINAL = 'hass_chris_final'
BASE_PATH = '/home/cmeier/code/data/thesis_results'

DT_RAW = 'raw'
DT_CH = 'changed'
DT_LF = 'last_fired'

#------------------------
# Configuration

MODEL_CLASS = BHMM
DATA_NAME = DN_HASS_CHRIS_FINAL
DATA_TYPE = DT_CH
TIME_DIFF = MIN30

#------------------------

DATASET_FILE_NAME = 'dataset_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF + ".joblib"
DATASET_FILE_PATH = BASE_PATH + '/' + DATA_NAME + '/datasets/' + DATASET_FILE_NAME

MODEL_NAME = 'model_' + MODEL_CLASS + '_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF
# todo this
#MODEL_FOLDER_PATH = BASE_PATH + '/' + DATA_NAME + '/models/' + MODEL_NAME
MODEL_FOLDER_PATH = BASE_PATH + '/' + DATA_NAME + '/best_model_selection/' + MODEL_NAME

MODEL_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME +".joblib"

MD_LOSS_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.loss.csv'
MD_LOSS_IMG_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.loss.png'
MD_INFST_IMG_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.inferred_states.png'
MD_CONF_MAT_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.confusion_matrix.numpy'
MD_METRICS_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.metrics.csv'
MD_ACT_DUR_DISTS_IMG_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.act_dur_dists.png'
MD_ACT_DUR_DISTS_DF_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.act_dur_dists.csv'
DATA_ACT_DUR_DISTS_IMG_FILE_PATH = MODEL_FOLDER_PATH + '/dataset.act_dur_dists.png'
DATA_ACT_DUR_DISTS_DF_FILE_PATH = MODEL_FOLDER_PATH + '/dataset.act_dur_dists.csv'


def load_domain_knowledge(file_path):
    with open(file_path, 'r') as file:
        import json
        data = json.load(file)
        act_data = data['activity_data']
        loc_data = data['loc_data']

    import datetime
    for act_data_point in act_data:
        act_data_point['start'] = datetime.datetime.strptime(
            act_data_point['start'],
            '%H:%M:%S').time()
        act_data_point['end'] = datetime.datetime.strptime(
            act_data_point['end'],
            '%H:%M:%S').time()
    return act_data, loc_data

def main():
    ctrl = Controller()
    ctrl.load_dataset_from_file(DATASET_FILE_PATH)

    # load model
    model_name = 'bhmm'
    ctrl.load_model(MODEL_FILE_PATH, model_name)
    ctrl.create_model_agnostics(model_name)
    hm = ctrl.get_model(model_name)._obs_lbl_hashmap
    file_path = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.lime_learning.png'
    file_paths = [file_path]
    labels_to_explain=['learning']
    import numpy as np
    # typical for activation pattern for dental care
    raw_feature_learning = np.array(
        [0, 0, 1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1], dtype=bool)
    ch_feature_learning = np.array(
        [0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0], dtype=bool)
    lf_feature_learning = np.array(
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0], dtype=bool)
    data = ctrl._dataset # type: _Dataset
    ctrl.plot_and_save_explanation(
        model_name,
        ch_feature_learning,
        labels_to_explain,
        file_paths
    )


if __name__ == '__main__':
    print('dataset filepath : ', DATASET_FILE_PATH)
    print('model filepath: ', MODEL_FILE_PATH)
    main()
    print('finished')


import plotly.figure_factory as ff
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from plotly.colors import n_colors
from datetime import timedelta

import numpy as np
from pyadlml.dataset.stats.devices import devices_on_off_stats, \
    devices_trigger_count, devices_dist, devices_trigger_time_diff, \
    device_tcorr, device_triggers_one_day


def ridge_line(df_act, t_range='day', n=1000):
    """
    https://plotly.com/python/violin/

    for one day plot the activity distribution over the day
    - sample uniform from each interval   
    """
    df = activities_dist(df_act.copy(), t_range, n)

  

    colors = n_colors('rgb(5, 200, 200)', 'rgb(200, 10, 10)', len(df.columns), colortype='rgb')
    data = df.values.T

    fig = go.Figure()
    i = 0
    for data_line, color in zip(data, colors):
        fig.add_trace(go.Violin(x=data_line, line_color=color, name=df.columns[i]))
        i += 1

    fig.update_traces(orientation='h', side='positive', width=3, points=False)
    fig.update_layout(xaxis_showgrid=False, xaxis_zeroline=False)
    return fig

def hist_counts(df_dev, scale_y='both'):
    """
    plots the trigger count of each device 
    """
    df = devices_trigger_count(df_dev.copy())
    df.reset_index(level=0, inplace=True)

    title = 'Count of on/off activations per Device'
    col_label = 'trigger count'
    col_device = 'device'
    df.columns = ['device', col_label]

    df = df.sort_values(by=col_label, axis=0, ascending=True)
    if scale_y == 'log':
        col_label_log = 'log '+ col_label
        df[col_label_log] = np.log(df[col_label])
        labels={col_label: col_label_log}
        return px.bar(df, x=col_label_log, y=col_device, \
                title=title, \
                labels=labels, orientation='h', height=450)

    elif scale_y == 'norm':
        labels={col_label: col_label}    
        return px.bar(df, x=col_label, y=col_device, \
                title=title, \
                labels=labels, orientation='h', height=450)

    elif scale_y == 'both':
        col_label_log = 'log '+col_label
        df[col_label_log] = np.log(df[col_label])
        labels={col_label: col_label_log}
        fig = go.Figure()

        #fig = px.bar(df, x=col_label, y=col_device, \
        #        title=title, \
        #        labels=labels, orientation='h', height=450)

        # Add Traces
        fig.add_trace(go.Bar(x=df[col_label], y=df[col_device], \
                                orientation='h', name='count'))
        fig.add_trace(go.Bar(x=df[col_label_log], y=df[col_device], \
                                orientation='h', name='log count', \
                                visible=False))

        # Add dropdown
        fig.update_layout(
            updatemenus=[
                dict(
                    buttons=list([
                        dict(
                            label = "no scaling", method = "update",
                            args = [{"visible": [True, False]},
                                {"title": title, "xaxis_title": 'asdf'}]),
                        dict(
                            label = "log scaled", method = "update", 
                            args = [{"visible": [False, True]},
                                {"title": title, "xaxis_title": 'asdf'}])
                    ]),
                    active=0,
                    direction="up",
                    showactive=True,
                    x=1.0,
                    xanchor="right",
                    y=-0.2,
                    yanchor="bottom"
                ),
            ]
        )
        fig.update_layout(
            title=title,
            xaxis_title="trigger count",
        )

        return fig
    else: 
        raise ValueError

def hist_trigger_time_diff(df_dev):
    """
        plots
    """
    df = devices_trigger_time_diff(df_dev.copy())
    fig = go.Figure()
    trace = go.Histogram(x=np.log(df['row_duration'].dt.total_seconds()/60),
                        nbinsx=200,
                      )
    fig.add_trace(trace)
    return fig

def hist_trigger_time_diff_separate(df_dev):
    """
    """
    df = devices_trigger_time_diff(df_dev.copy())
    df = df[['device','time_diff2']]

    dev_list = df.device.unique()
    num_cols = 2
    num_rows = int(np.ceil(len(dev_list)/2))
    fig = make_subplots(rows=num_rows, cols=num_cols)

    for i, device in enumerate(dev_list):
        dev_df = df[df.device  == device].dropna()['time_diff2']
        k = i%num_cols + 1
        j = int(i/num_cols) + 1
        fig.append_trace(
            go.Histogram(x=np.log(dev_df.dt.total_seconds()/60), 
            nbinsx=100, 
            name=device
            ), j,k
        )
    return fig

def hist_on_over_time(df_dev):
    """
        expectes device type 2
    """
    df = devices_dist(df_dev.copy(), t_range='day', n=1000)

    fig = go.Figure()
    # for every device add histogram to plot
    for col in df.columns:
        fig.add_trace(go.Histogram(x=df[col], nbinsx=50, name=col))

    # Overlay both histograms
    fig.update_layout(barmode='overlay')

    # Reduce opacity to see both histograms
    fig.update_traces(opacity=0.75)
    return fig


def heatmap_trigger_one_day(df_dev, t_res='1h'):
    """
    computes the heatmap for one day where all the device triggers are showed
    """
    if type(t_res) == str:
        df = device_triggers_one_day(df_dev.copy(), t_res)

        #plot stuff
        fig = go.Figure(data=go.Heatmap(
                z=df.T.values,
                x=df.index,
                y=df.columns,
                colorscale='Viridis'))

        fig.update_layout(
            title='Triggers one day',
            xaxis_nticks=24)
        return fig
    else:
        # Create figure
        fig = go.Figure()

        # create for each timeframe the data
        # Add traces, one for each slider step
        for t_val in t_res:
            df = device_triggers_one_day(df_dev.copy(), t_val)
            fig.add_trace(
                go.Heatmap(
                        z=df.T.values,
                        x=df.index,
                        y=df.columns,
                        colorscale='Viridis'))

        # Make 0th trace visible
        fig.data[0].visible = True

        # Create and add slider
        steps = []
        for i in range(len(fig.data)):
            step = dict(
                label=t_res[i],
                method="restyle",
                args=[{"visible": [False] * len(fig.data)},
                    {"title": "Slider switched to step: " + str(i)}],  # layout attribute
            )
            step["args"][0]["visible"][i] = True  # Toggle i'th trace to "visible"
            steps.append(step)

        sliders = [dict(
            active=0,
            currentvalue={'visible':False},
            pad={"t": 70},
            steps=steps
        )]

        fig.update_xaxes(patch=dict(nticks=24))

        fig.update_layout(
            title='Triggers one day',
            sliders=sliders
        )
        return fig

def heatmap_time_diff(df_dev, t_windows):
    x_label = 'Devices'
    y_label = x_label
    color = 'trigger count'
    name = 'Device triggers'

    # get the list of cross tabulations per t_window
    lst = device_tcorr(df_dev, t_windows)

    if len(t_windows) == 1:
        # make a single plot
        df = lst[0]
        fig = go.Figure(data=go.Heatmap(
        name=name,
        #labels=dict(x=x_label, y=y_label, color=color),
        z=df,
        x=df.columns,
        y=df.index,
        hoverongaps = False))
        return fig
    else:
        # make multiple subplots
        num_cols = 2
        num_rows = int(np.ceil(len(lst)/2))
        print('nr: ', num_rows)
        fig = make_subplots(rows=num_rows, cols=num_cols)

        for i, device in enumerate(lst):
            df = lst[i]
            col = i%num_cols + 1
            row = int(i/num_cols) + 1
            fig.append_trace(
                go.Heatmap(
                name=name,
                #labels=dict(x=x_label, y=y_label, color=color),
                z=df,
                x=df.columns,
                y=df.index,
                hoverongaps = False
                ),col,row
            )
        return fig


def hist_on_off(df_dev):
    """
        plots the fraction a device is on vs off over the whole time
    """
    title='Devices fraction on/off'

    df = devices_on_off_stats(df_dev)
    df = df.sort_values(by='frac_on', axis=0)

    y = df.index

    fig = go.Figure()
    fig.add_trace(go.Bar(
        y=y,
        x=df['frac_on'],
        name='on',
        orientation='h',
        marker=dict(
            color='rgba(1, 144, 105, 0.8)',
        )
    ))
    fig.add_trace(go.Bar(
        y=y,
        x=df['frac_off'],
        name='off',
        orientation='h',
        marker=dict(
            color='rgba(58, 71, 80, 0.8)',
        )
    ))

    fig.update_layout(barmode='stack', title=title)
    return fig

def hist_trigger_time_diff(df_dev):
    """
        plots
    """
    n_bins = 100
    title='Time difference between device triggers'
    log_sec_col = 'total_log_secs'
    sec_col = 'total_secs'
    df = devices_trigger_time_diff(df_dev.copy())
    
    # convert timedelta to total minutes
    df[sec_col] = df['row_duration']/timedelta(seconds=1)
    df[log_sec_col] = np.log(df[sec_col])
    X = np.log(df[sec_col]).values[:-1]

    #return df
    fig = make_subplots(specs=[[{"secondary_y": True}]])
    
    #hover_data = {'sec: ', np.random.random(len(df))}
    hover_template = "count: %{y:}<br>bin: %{x:} </br><extra></extra>"
    fig.add_trace(go.Histogram(x=df[log_sec_col].values[:-1],
                        name='# device triggers',
                        nbinsx=n_bins,
                        hovertemplate=hover_template,),
                        secondary_y=False
                        )
    
    hist, bin_edges = np.histogram(df[log_sec_col].values[:-1], n_bins)
    mask = (np.array([0,1]*int((n_bins/2))+[0]) == 1) # select every scnd element
    left_bins = np.asarray([0]+list(bin_edges[mask]))
    
    hist, bin_edges = np.histogram(X, n_bins)
    left_bins = bin_edges[:-1]
    cum_percentage = hist.cumsum()/hist.sum()
    
    fig.update_layout(
            title=title,
            xaxis_title="log min",
            bargap=0.1,
        )
    fig.add_trace(go.Scatter(x=left_bins, 
                            y=cum_percentage, 
                            name="percentage left"),
                    secondary_y=True)

    # Set y-axes titles
    fig.update_yaxes(title_text="count", secondary_y=False)
    fig.update_yaxes(title_text="%", secondary_y=True)
    return fig